package main

import (
	"bytes"
	"crypto/tls"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"net/http"
	"net/url"
	"strings"
	"sync"

	"github.com/PuerkitoBio/goquery"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/viper"
	"github.com/xlzd/gotp"
)

const (
	namespace = "seafile" // Prefix for Prometheus metrics.
)

// Exporter collects seafile stats from the given URI and exports them using
// the prometheus metrics package.
type Exporter struct {
	config *Config
	mutex  sync.RWMutex
	client *http.Client

	server_info *SeafileServerInfo

	up                        prometheus.Gauge
	system_info               *prometheus.GaugeVec
	update_available          *prometheus.GaugeVec
	license                   *prometheus.GaugeVec
	license_maxusers          prometheus.Gauge
	users                     prometheus.Gauge
	users_active              prometheus.Gauge
	repos                     prometheus.Gauge
	total_files               prometheus.Gauge
	groups                    prometheus.Gauge
	multi_tenancy_enabled     prometheus.Gauge
	is_pro                    prometheus.Gauge
	total_storage             prometheus.Gauge
	total_devices             prometheus.Gauge
	current_connected_devices prometheus.Gauge
}

type Config struct {
	listen_address        string
	metrics_path          string
	ignore_ssl            bool
	seafile_scrape_uri    string
	seafile_auth_user     string
	seafile_auth_password string
	seafile_auth_secret   string
	seafile_changelog     string

	Token *SeafileAuthToken
}

type SeafileAuthRequest struct {
	Username string `json:"username"`
	Password string `json:"password"`
}

type SeafileAuthToken struct {
	Token string `json:"token"`
}

type SeafileServerInfo struct {
	Version string `json:"version"`
}

type SeafileSystemInfo struct {
	UsersCount                   int    `json:"users_count"`
	ActiveUsersCount             int    `json:"active_users_count"`
	ReposCount                   int    `json:"repos_count"`
	TotalFilesCount              int    `json:"total_files_count"`
	GroupsCount                  int    `json:"groups_count"`
	OrgCount                     int    `json:"org_count"`
	MultiTenancyEnabled          bool   `json:"multi_tenancy_enabled"`
	IsPro                        bool   `json:"is_pro"`
	WithLicense                  bool   `json:"with_license"`
	LicenseExpiration            string `json:"license_expiration"`
	LicenseMode                  string `json:"license_mode"`
	LicenseMaxusers              int    `json:"license_maxusers"`
	LicenseTo                    string `json:"license_to"`
	TotalStorage                 int    `json:"total_storage"`
	TotalDevicesCount            int    `json:"total_devices_count"`
	CurrentConnectedDevicesCount int    `json:"current_connected_devices_count"`
}

type SeafileReleaseInfo struct {
	Version     string
	ReleaseDate string
}

func FetchStatus(e *Exporter) (*SeafileServerInfo, error) {
	url := e.config.seafile_scrape_uri + "/api2/server-info/"
	log.Infof("Fetch status from %s", url)
	req, err := http.NewRequest("GET", url, nil)
	if err != nil {
		log.Errorf("Unable to create request to %s: %v", url, err)
		return nil, err
	}

	resp, err := e.client.Do(req)
	if err != nil {
		log.Error("Unable to fetch seafile status")
		return nil, err
	}
	switch resp.StatusCode {
	case 200:
	case 401:
		return nil, errors.New("authentication with seafile failed")
	default:
		return nil, fmt.Errorf("seafile returned %s", resp.Status)
	}

	seafile_server_info := SeafileServerInfo{}
	if err := json.NewDecoder(resp.Body).Decode(&seafile_server_info); err != nil {
		defer resp.Body.Close()
		return nil, err
	} else {
		defer resp.Body.Close()
	}
	return &seafile_server_info, nil
}

func FetchSystemInfo(e *Exporter, fetchToken bool) (*SeafileSystemInfo, error) {
	if fetchToken || e.config.Token == nil {
		token, err := e.FetchSeafileToken()
		if err != nil {
			log.Errorf("Unable to determine token: %v", err)
			return nil, err
		}
		e.config.Token = token
	}

	url := e.config.seafile_scrape_uri + "/api/v2.1/admin/sysinfo/"
	log.Infof("Fetch status from %s", url)
	req, err := http.NewRequest("GET", url, nil)
	if err != nil {
		log.Errorf("Unable to create request to %s: %v", url, err)
		return nil, err
	}

	req.Header.Set("Authorization", e.config.Token.getHeader())

	resp, err := e.client.Do(req)
	if err != nil {
		log.Error("Unable to fetch seafile status")
		return nil, err
	}
	switch resp.StatusCode {
	case 200:
	case 401:
		if fetchToken {
			return nil, errors.New("authentication with seafile failed")
		} else {
			return FetchSystemInfo(e, true)
		}
	default:
		return nil, fmt.Errorf("seafile returned %s", resp.Status)
	}

	seafile_system_info := SeafileSystemInfo{}
	if err := json.NewDecoder(resp.Body).Decode(&seafile_system_info); err != nil {
		defer resp.Body.Close()
		return nil, err
	} else {
		defer resp.Body.Close()
	}
	return &seafile_system_info, nil
}

func FetchSeafileRelease(e *Exporter) (*SeafileReleaseInfo, error) {
	req, err := http.NewRequest("GET", e.config.seafile_changelog, nil)
	if err != nil {
		log.Errorf("Unable to create request: %v", err)
		return nil, err
	}

	req.Header.Set("User-Agent", "Seafile-Exporter")
	resp, err := e.client.Do(req)
	if err != nil {
		log.Error("Unable to fetch seafile update status")
		return nil, err
	}

	switch resp.StatusCode {
	case 200:
	default:
		return nil, fmt.Errorf("seafile returned %s", resp.Status)
	}

	doc, err := goquery.NewDocumentFromReader(resp.Body)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	latestVersionTag := doc.Find("h3").First()
	latestVersion := latestVersionTag.Text()
	versionInfo := strings.Split(latestVersion, " ")

	release_info := SeafileReleaseInfo{
		Version:     versionInfo[0],
		ReleaseDate: versionInfo[1],
	}
	return &release_info, nil
}

// ParseConfig parse exporter binary options from command line
func ParseConfig() *Config {
	flag.String("conf", "./config.toml", "Configuration file for exporter")
	flag.Parse()

	v := viper.New()

	// Provide all configurations as environment variable as well.
	v.SetEnvPrefix(strings.ToUpper(namespace))
	v.AutomaticEnv()

	v.SetDefault("listen_address", "0.0.0.0:9388")
	v.SetDefault("metrics_path", "/metrics")
	v.SetDefault("ignore_ssl", false)
	v.SetDefault("seafile_scrape_uri", "http://localhost/")
	v.SetDefault("seafile_auth_user", "admin")
	v.SetDefault("seafile_auth_password", "seafile")
	v.SetDefault("seafile_auth_secret", "")
	v.SetDefault("seafile_changelog", "https://manual.seafile.com/changelog/changelog-for-seafile-professional-server/")
	v.SetConfigFile(flag.Lookup("conf").Value.String())
	v.SetConfigType("toml")

	err := v.ReadInConfig() // Find and read the config file
	if err != nil {         // Handle errors reading the config file
		log.Printf("Error reading config file: %s. Using defaults.", err)
	}

	return &Config{
		listen_address:        v.GetString("listen_address"),
		metrics_path:          v.GetString("metrics_path"),
		ignore_ssl:            v.GetBool("ignore_ssl"),
		seafile_scrape_uri:    v.GetString("seafile_scrape_uri"),
		seafile_auth_user:     v.GetString("seafile_auth_user"),
		seafile_auth_password: v.GetString("seafile_auth_password"),
		seafile_auth_secret:   v.GetString("seafile_auth_secret"),
		seafile_changelog:     v.GetString("seafile_changelog"),
		Token:                 nil,
	}
}

// Returns an initialized Exporter.
func NewExporter(c *Config) (*Exporter, error) {

	return &Exporter{
		config: c,
		client: &http.Client{
			Transport: &http.Transport{
				TLSClientConfig: &tls.Config{InsecureSkipVerify: c.ignore_ssl},
			},
		},
		server_info: nil,
		up: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "up",
			Help:      "Indicates if the metrics could be scraped by the exporter.",
		}),
		system_info: prometheus.NewGaugeVec(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "system_info",
			Help:      "Seafile server info",
		},
			[]string{"version"},
		),
		update_available: prometheus.NewGaugeVec(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "update_available",
			Help:      "Seafile server update",
		},
			[]string{"version", "available_version"},
		),
		license: prometheus.NewGaugeVec(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "license",
			Help:      "Seafile license information",
		},
			[]string{"license_expiration", "license_to", "license_mode"},
		),
		license_maxusers: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "license_maxusers",
			Help:      "Seafile license: maximum users",
		}),

		users: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "users",
			Help:      "Seafile system info: users",
		}),
		users_active: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "users_active",
			Help:      "Seafile system info: users_active",
		}),
		repos: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "repos",
			Help:      "Seafile system info: repos",
		}),
		total_files: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "total_files",
			Help:      "Seafile system info: total_files",
		}),
		groups: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "groups",
			Help:      "Seafile system info: groups",
		}),
		multi_tenancy_enabled: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "multi_tenancy_enabled",
			Help:      "Seafile system info: multi_tenancy_enabled",
		}),
		is_pro: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "is_pro",
			Help:      "Seafile system info: is_pro",
		}),
		total_storage: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "total_storage",
			Help:      "Seafile system info: total_storage",
		}),
		total_devices: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "total_devices",
			Help:      "Seafile system info: total_devices",
		}),
		current_connected_devices: prometheus.NewGauge(prometheus.GaugeOpts{
			Namespace: namespace,
			Name:      "current_connected_devices",
			Help:      "Seafile system info: current_connected_devices",
		}),
	}, nil
}

func (e *Exporter) FetchSeafileToken() (*SeafileAuthToken, error) {
	request_url, err := url.JoinPath(e.config.seafile_scrape_uri, "/api2/auth-token/")
	if err != nil {
		return nil, err
	}

	m := SeafileAuthRequest{
		Username: e.config.seafile_auth_user,
		Password: e.config.seafile_auth_password,
	}
	b, err := json.Marshal(m)
	if err != nil {
		log.Errorf("Unable to create request to %s: %v", request_url, err)
	}

	req, err := http.NewRequest("POST", request_url, bytes.NewReader(b))
	if err != nil {
		log.Errorf("Unable to create request to %s: %v", request_url, err)
	}

	req.Header.Set("Accept", "application/json")
	req.Header.Set("Content-Type", "application/json")
	if e.config.seafile_auth_secret != "" {
		totp := gotp.NewDefaultTOTP(e.config.seafile_auth_secret)
		req.Header.Set("x-seafile-otp", totp.Now())
	}

	resp, err := e.client.Do(req)
	if err != nil {
		log.Error("Unable to fetch seafile status")
		return nil, err
	}
	switch resp.StatusCode {
	case 200:
	case 401:
		return nil, errors.New("authentication with seafile failed")
	default:
		return nil, fmt.Errorf("seafile returned %s", resp.Status)
	}

	token := SeafileAuthToken{}
	if err := json.NewDecoder(resp.Body).Decode(&token); err != nil {
		defer resp.Body.Close()
		return nil, err
	} else {
		defer resp.Body.Close()
	}
	return &token, nil
}

func (e *SeafileAuthToken) getHeader() string {
	return "Bearer " + e.Token
}

func (e *Exporter) readMetrics(status *SeafileServerInfo) error {
	e.server_info = status
	e.system_info.With(
		prometheus.Labels{
			"version": status.Version,
		}).Set(1)

	return nil
}

func (e *Exporter) readReleaseInfo(release *SeafileReleaseInfo) error {
	log.Infof("Latest available version: %s", release.Version)
	current_version := e.server_info.Version

	version_available := 0.0
	if release.Version > current_version {
		version_available = 1.0
	}

	e.update_available.With(
		prometheus.Labels{
			"version":           current_version,
			"available_version": release.Version,
		}).Set(version_available)

	return nil
}

func (e *Exporter) readSystemInfo(system *SeafileSystemInfo) error {
	e.license_maxusers.Set(float64(system.LicenseMaxusers))

	with_license := 0.0
	if system.WithLicense {
		with_license = 1.0
	}
	multi_tenancy_enabled := 0.0
	if system.MultiTenancyEnabled {
		multi_tenancy_enabled = 1.0
	}
	is_pro := 0.0
	if system.IsPro {
		is_pro = 1.0
	}

	e.license.With(
		prometheus.Labels{
			"license_expiration": system.LicenseExpiration,
			"license_to":         system.LicenseTo,
			"license_mode":       system.LicenseMode,
		}).Set(with_license)

	e.users.Set(float64(system.UsersCount))
	e.users_active.Set(float64(system.ActiveUsersCount))
	e.repos.Set(float64(system.ReposCount))
	e.total_files.Set(float64(system.TotalFilesCount))
	e.groups.Set(float64(system.GroupsCount))
	e.multi_tenancy_enabled.Set(multi_tenancy_enabled)
	e.is_pro.Set(is_pro)
	e.total_storage.Set(float64(system.TotalStorage))
	e.total_devices.Set(float64(system.TotalDevicesCount))
	e.current_connected_devices.Set(float64(system.CurrentConnectedDevicesCount))

	return nil
}

// Describe describes all the metrics ever exported by the monit exporter. It
// implements prometheus.Collector.
func (e *Exporter) Describe(ch chan<- *prometheus.Desc) {
	e.up.Describe(ch)
	e.system_info.Describe(ch)
	e.license.Describe(ch)
	e.license_maxusers.Describe(ch)
}

func (e *Exporter) scrape() error {
	data, err := FetchStatus(e)
	if err == nil {
		return e.readMetrics(data)
	}
	return err
}

func (e *Exporter) scrapeSystem() error {
	data, err := FetchSystemInfo(e, false)
	if err == nil {
		return e.readSystemInfo(data)
	}
	return err
}

func (e *Exporter) scrapeUpdate() error {
	data, err := FetchSeafileRelease(e)
	if err == nil {
		return e.readReleaseInfo(data)
	}
	return err
}

// Collect fetches the stats from configured monit location and delivers them
// as Prometheus metrics. It implements prometheus.Collector.
func (e *Exporter) Collect(ch chan<- prometheus.Metric) {
	e.mutex.Lock() // Protect metrics from concurrent collects.
	defer e.mutex.Unlock()
	if err := e.scrape(); err != nil {
		e.up.Set(0)
		log.Errorf("Error getting seafile status: %v", err)
	} else {
		if err := e.scrapeSystem(); err != nil {
			e.up.Set(0)
			log.Errorf("Error getting seafile update status: %v", err)
		} else {
			if err := e.scrapeUpdate(); err != nil {
				e.up.Set(0)
				log.Errorf("Error getting seafile update status: %v", err)
			} else {
				e.up.Set(1)
			}
		}
	}
	e.up.Collect(ch)
	e.system_info.Collect(ch)
	e.update_available.Collect(ch)
	e.license.Collect(ch)
	e.license_maxusers.Collect(ch)
	e.users.Collect(ch)
	e.users_active.Collect(ch)
	e.repos.Collect(ch)
	e.total_files.Collect(ch)
	e.groups.Collect(ch)
	e.multi_tenancy_enabled.Collect(ch)
	e.is_pro.Collect(ch)
	e.total_storage.Collect(ch)
	e.total_devices.Collect(ch)
	e.current_connected_devices.Collect(ch)
}

func main() {

	config := ParseConfig()
	exporter, err := NewExporter(config)

	if err != nil {
		log.Fatal(err)
	}
	prometheus.MustRegister(exporter)

	log.Printf("Starting seafile-exporter: %s", config.listen_address)
	http.Handle(config.metrics_path, promhttp.Handler())
	http.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		_, err := w.Write([]byte(`<html>
			<head><title>Seafile exporter</title></head>
			<body>
			<h1>Seafile exporter</h1>
			<p><a href="` + config.metrics_path + `">Metrics</a></p>
			</body>
			</html>`))

		if err != nil {
			log.Fatal(err)
		}
	})

	log.Fatal(http.ListenAndServe(config.listen_address, nil))
}
